/* ***** BEGIN LICENSE BLOCK *****
 * Licensed under Version: MPL 1.1/GPL 2.0/LGPL 2.1
 * Full Terms at http://mozile.mozdev.org/license.html
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Playsophy code.
 *
 * The Initial Developer of the Original Code is Playsophy
 * Portions created by the Initial Developer are Copyright (C) 2002-2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *	Karl Guertin <grayrest@grayrest.com>
 *	James A. Overton <james@overton.ca>
 *	Christian Stocker (Bitflux)
 *	Paul Everitt
 *	David Palm
 *	Richard Prescott
 *
 * ***** END LICENSE BLOCK ***** */

/**
@fileoverview

Parse MES

This code is responsible for parsing the special Mozile Editing Scheme XML
files. There's nothing very fancy going on here, but it is a lot of work.
*/

/**
Create MES data structure.

@constructor
*/
function MozileMES( images, commands, tags, toolbar, eDOM ) {
	this.images = images || new Array();
	this.commands = commands || new Array();
	this.tags = tags || new Array();
	this.toolbar = toolbar || new Array();
	this.eDOM = eDOM || new Object();
}

var mozileMES = new Object();
//Global methods
mozileMES.loadMES = mozileLoadMES;
mozileMES.findByName = mozileFindByName;
mozileMES.getCurrentMES = mozileGetCurrentMES;
mozileMES.getMESIndex = mozileGetMESIndex;

// Internal MES available.
mozileMES.internal = [
	"mes-XHTML-en-US.xml",
	"mes-XHTML-en-UK.xml",
	"mes-XHTML-text.xml",
	"mes-XHTML-test.xml"
];

function mozileGetMESIndex( mesName ) {
	for (var i=0; i<mozileMES.internal.length; i++ ) {
		if ( mozileMES.internal[i] == mesName ) {
			return i;
		}
	}
	return null;
}

/**
 Gets the current MES using the following order of precedence:
 1. RSD
 2. External
 3. Internal
 4. XHTML default.

*/
function mozileGetCurrentMES( target ) {
	var f = ["mozileParseMES.js","mozileGetCurrentMES"];

	try {
	// 1. mes from RSD
		var mes = target.document.MOZILE_CURRENT_CONFIG.mes;
		if ( mes && mes!="" ) {
			return mes;
		}

	// 2. External custom MES
		if ( mozilePrefs.getBoolPref("mozile.MESUseExternalBool") ) {
			mes = mozilePrefs.getCharPref("mozile.MESExternalString");
			if ( mes && mes!="" ) {
				return mes;
			}
		}
		
	// 3. Internal MES
		mes = mozilePrefs.getCharPref("mozile.MESDefaultString");
		if ( mes && mes!="" ) {
			if ( mozileGetMESIndex( mes ) !== null ) {
				return MOZILEC.MES_DIR + mes;
			}
		}
	} catch (e) {
		mozileDebug(f,1,"Error getting current MES: "+e);
	}
	// 4. Fail safe default.
	return "XHTML";
}

/**
Load and parse the MES configuration file.
Amongst other things this defines the commands and toolbar buttons.

<p>The toolbar data is stored in the target document:<br />
<code>mozileInjectTarget.document.mozileMES</code>

@param {Window} mozileInjectTarget target window being edited.
@param {String} mes MES document to parse.
*/
function mozileLoadMES(mozileInjectTarget, mes) {
	var f = ["mozileParseMES.js","mozileLoadMES"];

	var MES, MES_doc, mesloaded;

	if ( !mes || mes=="" || mes=="XHTML" ) {
		mozileDebug(f,3,"Loading XHTML");
		MES_doc = MOZILEC.MES_DEFAULT;
	} else {
		MES_doc = mes;
	}
	mozileDebug(f,3,"MES to load: "+ MES_doc);
	
	mozileInjectTarget.document.mozileMES = new MozileMES();
	var mozileMESi = mozileInjectTarget.document.mozileMES;

	mesloaded = false;
	try {
  	MES = mozileInjectTarget.document.implementation.createDocument("", "", null);
		MES.async = false;
		mesloaded = MES.load(MES_doc);
		if(!mesloaded) {
			mozileDebug(f,1,"Error loading MES: "+MES_doc);
			MES=null;		
		}
	}
	catch(e)
	{
		mozileDebug(f,1,"Exception loading MES: "+e);
		MES = null;
	}

	//If error or XML parsing error load the fail-safe default MES...
	//(If RSD or External MES could try user default before fail-safe default.)
	if ( MES==null || !MES.documentElement || MES.documentElement.tagName == "parsererror") {
		mozileDebug(f,2,"Could not load MES");
		if ( MES_doc !== MOZILEC.MES_DEFAULT ) {
			mozileLoadMES(mozileInjectTarget,"XHTML");
		}
		return;
	}
	else {	// success!	
		try {
			mozileMESi.images = mozileParseIcons(MES);
			mozileMESi.commands = mozileParseCommands(MES);
			mozileMESi.tags = mozileParseTags(MES);
			mozileMESi.toolbar = mozileParseToolbar(MES);
			mozileMESi.eDOM = mozileParseEDOM(MES);

			mozileParseSpecialChars(mozileInjectTarget, MES);

			mozileDebug(f,3,"MES update completed.");
		} catch(e) {
			mozileDebug(f,1,"Exception updating MES: "+e);
		}
	}
}
/**
Parse fixed toolbar command elements.
Looks for fixed toolbar items to be disabled.

<p>Example command element format:<br />
 &lt;command name="bug" disabled="true" &gt;

@return array of names as strings.
@type Array
*/
function mozileParseToolbar(MES) {
	var f = ["mozileParseMES.js","mozileParseToolbar"];
	var id = new Array();
 	try{
		var commands = MES.documentElement.getElementsByTagName("commands");
		if (commands.length > 0) {
			var command = commands[0].getElementsByTagName("command");
			for( var i=0; i<command.length; i++ ) {
				if ( command[i].getAttribute("type") == "toolbar" ) {
					var name = command[i].getAttribute("name");
					var disabled = command[i].getAttribute("disabled");
					if ( name && disabled=="true" ) {
						if (mozileGetToolbarById("mozile-"+name)) {
							id.push(name);
						}
					}
				}
			}
		}
	}catch(e){
		mozileDebug(f,1,"Exception parsing MES fixed toolbar command: "+e);
	}
	return id;
}

/**
Parse eDOM part of the MES.
Not strict parsing or validation of the xhtml value, so get it right...

@return xhtml default container.
@type Object.
*/
function mozileParseEDOM(MES) {
	var f = ["mozileParseMES.js","mozileParseEDOM"];
	
	var mesEDOM = new Object();
	mesEDOM.xhtmlDefaultContainer = "div";
	try {
		var dc = MES.getElementsByTagName("container");
		if (dc.length>0 && dc[0].hasAttribute("default")) {
			var c = dc[0].getAttribute("default");
			//Could check is a valid xhtml element.
			mesEDOM.xhtmlDefaultContainer = c;
		}
	} catch(e) {
		mozileDebug(f,1,"Exception parsing eDOM MES: "+e);
	}
	return mesEDOM;
}

function mozileParseIcons(MES) {
	var icons = MES.documentElement.getElementsByTagName("icons")[0];
	var images = new Array();
	images[0] = icons.getAttribute("small");
	images[1] = icons.getAttribute("large");
	images[2] = icons.getAttribute("format");
	return images;
}

function mozileParseCommands(MES) {
	var commands = MES.documentElement.getElementsByTagName("commands")[0];
	var commandTags = commands.childNodes;
	var commandTag, parsed, i;
	var commandArray = new Array();

	for(i=0; i < commandTags.length; i++) {
		commandTag=commandTags[i];
		if( (commandTag.nodeType == Node.ELEMENT_NODE) && (commandTag.tagName == "command") ) {
			parsed = parseCommand(commandTag);
			if(parsed) {
				commandArray.push( parsed );
			}
		}
	}

 //This is a private function of mozileParseCommands.
 function parseCommand(tag) {
 	var f = ["mozileParseMES.js","parseCommand"];
	var carr = new Array();
	var type = tag.getAttribute("type");
		
 	// If fixed toolbar item, ignore it.
	if (type == "toolbar")
		return null;

		// deal with common attributes
	carr[0] = tag.getAttribute("name");
	carr[1] = tag.getAttribute("title");
	carr[2] = tag.getAttribute("tooltip");
	carr[3] = tag.getAttribute("icon");
	carr[4] = tag.getAttribute("modifiers");
	carr[5] = tag.getAttribute("key");
	carr[6] = tag.getAttribute("type");
	
		// deal with special attributes
	var args=new Array();

	switch(type)
	{
		// style is something specific to XHTML
		// no namespace is involved
		case "toggleStyle":
			args[0] = tag.getAttribute("property");
			args[1] = tag.getAttribute("value");
			args[2] = tag.getAttribute("default");
			args[3] = tag.getAttribute("activeproperty");
			//
			var re = /\s*,\s+/;
			args[0] = args[0].split(re,10);
			args[1] = args[1].split(re,10);
			// default value may be optional at some time in the future...
			if ( args[2] && args[2]!="" ) {
				args[2] = args[2].split(re,10);
			}else{
				args[2]=new Array();
			}
			//Number of properties used when updating active toolbar.
			if ( !args[3] ) {
				args[3] = args[0].length;
			}
			break;
			
		// XXX: toggleAttribute doesn't care about attribute's namespace
		// it is not a breaking issue for now  - Richard
		case "toggleAttribute":
			args[0] = tag.getAttribute("attribute");
			args[1] = tag.getAttribute("value");
			args[2] = tag.getAttribute("default");
			break;

		// style is something specific to XHTML
		// no namespace is involved
		case "styleText":
			args[0] = tag.getAttribute("property");
			args[1] = tag.getAttribute("value");
			break;

		// style is something specific to XHTML
		// no namespace is involved
		case "styleLines":
			args[0] = tag.getAttribute("property");
			args[1] = tag.getAttribute("value");
			break;

		case "setLinesContainer":
			args[0] = tag.getAttribute("namespace");
			args[1] = tag.getAttribute("tag");
			break;

		case "wrapText":
			args[0] = tag.getAttribute("namespace");
			args[1] = tag.getAttribute("tag");
			break;

		// XHTML ul and ol only. 
		case "toggleListLines":
			args[0] = tag.getAttribute("tag");
			args[1] = tag.getAttribute("old");
			break;

		case "indent":
		case "outdent":
			args[0] = tag.getAttribute("value");
			break;

		case "indentLists":
		case "outdentLists":
		case 'link':
		case "unlink":
		case "unformat":
		case 'table':
		case 'image':
		case 'hr':
		case 'no-op':
			break;

		case 'x-element':
			args[0] = tag.getAttribute("tag");
			break;
			
		case "reload":
			args[0] = tag.getAttribute("contenttype");
			break;

		case 'mozileCommand':
			args[0] = tag.getAttribute("commandname");
			break;

			// if the command type hasn't been found, throw an error
		default:
			mozileDebug(f,1,'Command type "'+ type +'" not defined!');
			return null;
	}
	
	carr[7] = args;
		// done
	return carr;
 }

 return commandArray;
} //End of private functions.

/**
Parse the contents of the first "tags" element in the MES.

*/
/*
	GENERAL FORMAT
	tag
		   [use] {String}   tag name to reuse.
		0  name {String}    Note: Can't have duplicate name.
		1  title {String}
		2  unique {String}  "true"|"false" [Can be left out, in which case it defaults to false.]
		3  command array
		4  key array        Keys not in use yet.

	<tags>
		<tag name="pre" title="Pre-formatted" unique="false">
			<commands>
				<commandGroup name="format" title="Format " tooltip="Format lines" accesskey="o" button="true" >
					<command name="p" accesskey="P"/>
					<command name="h1" accesskey="1"/>
				</commandGroup>
				<command name="separator"/>
				<command name="bold" accesskey="B" button="true"/>
			</commands>
		</tag>

		<tag name="p" title="Paragraph" unique="false">
			<commands use="pre"/>
			<commands>
				<command name="separator"/>
				<command name="image" accesskey="m" button="true"/>
				<command name="table" accesskey="T" button="true"/>
			</commands>
		</tag>

		<tag name="h1" title="Heading 1" use="p"/>

	</tags>

*/
function mozileParseTags(MES) {
	var tags = MES.documentElement.getElementsByTagName("tags")[0];
	var tagsChildNodes = tags.childNodes;
	var parsed, tagsChild, i, j;
	var tagArray = new Array();
	
	for(j=0; j < tagsChildNodes.length; j++) {
		tagsChild=tagsChildNodes[j];
		if( (tagsChild.nodeType == Node.ELEMENT_NODE) && (tagsChild.tagName == "tag") ) {
			parsed = parseTag(tagsChild, tagArray)
			if(parsed) {
				tagArray.push( parsed );
			}
		}
	}
// NB : functions below are private to mozileParseTags.

/**
The "use" attribute specifies commands/keys to reuse. These are block copied to the command/key arrays. 
Individual keys/commands that are alrady present in the array are over written.
Those not in the array are appended to the end of the array.

@param {Element} tag
@param {Array} tagArray array of previously parsed tag elements.
@return tarr - array containing tag, command and key data; 0 if no valid data.
*/
function parseTag(tag, tagArray) {
	var tarr = new Array();
	var use, useIndex;
		
	// deal with common attributes
	tarr[0] = tag.getAttribute("name");
	tarr[1] = tag.getAttribute("title");
	tarr[2] = tag.getAttribute("unique");
	
	//alert("Parsing Tag Tag <"+tarr[0]+">");
	// check for duplicates
	if(mozileMES.findByName(tarr[0],tagArray)>=0) {
		alert("Tag "+ tarr[0] +" is defined more than once!");
		return 0;
	}
	
	// if we are copying a whole tag, then find it and copy it, then return
	use = tag.getAttribute("use");
	if(use && use!="") {
		useIndex=mozileMES.findByName(use, tagArray);
		//alert("Reuse all from "+ use + " at " +useIndex);
		if(useIndex >= 0) {
			tarr[3]=[].concat(tagArray[useIndex][3]);
			tarr[4]=[].concat(tagArray[useIndex][4]);
			return tarr;
		} else {
			alert("The "+tarr[0]+" tag tells me to 'use' the commands and keys from "+use+" tag, but they are not yet defined!");
		}
	}
	
	//Copy any "commands" or "keys" that have a "use" attribute.
	var children = tag.childNodes;
	var commandTags = new Array();
	var commands = new Array();
	var reuseCommands = 0;
	var keyTags = new Array();
	var keys = new Array();
	var reuseKeys = 0;
	for(var i=0; i<children.length; i++) {
		if( (children[i].nodeType == Node.ELEMENT_NODE) && (children[i].tagName == "commands") ) {
			commandTags = children[i].childNodes;
			use = children[i].getAttribute("use");
			if(use && use!="") {
				useIndex=mozileMES.findByName(use,tagArray);		// get the index of the tag name
				if(useIndex >= 0) {
					commands=commands.concat(tagArray[useIndex][3]);	// the third entry holds commands
					reuseCommands=1;
				} else {
					alert("The "+tarr[0]+" tag tells me to 'use' the commands from "+use+" tag, but they are not yet defined!");
				}
			}
		}
		if( (children[i].nodeType == Node.ELEMENT_NODE) && (children[i].tagName == "keys") ) {
			keyTags = children[i].childNodes;
			use = children[i].getAttribute("use");
			if(use && use!="") {
				useIndex=mozileMES.findByName(use,tagArray);		// get the index of the tag name
				if(useIndex >= 0) {
					keys=keys.concat(tagArray[useIndex][4]);	    // the fourth entry holds keys
					reuseKeys=1;
				} else {
					alert("The "+tarr[0]+" tag tells me to 'use' the keys from "+use+" tag, but they are not yet defined!");
				}
			}
		}
	}
	
	//Parse all the command tags into the commands array
	var commandIndex;
	for(var j=0; j<commandTags.length; j++) {
		tag = commandTags[j];
		if ( tag.nodeType == Node.ELEMENT_NODE ) {
  		if(reuseCommands) {
  			if(tag.tagName=="command") {
					if ( tag.getAttribute("name") == "separator" ) {
					//Always add the toolbar separator.
  					commands.push( parseTagCommand(tag) );
					}else {
	  				commandIndex = mozileMES.findByName(tag.getAttribute("name"), commands);
	  				if(commandIndex >= 0) { //If command already exits, overwrite it.
	  					commands[commandIndex] = parseTagCommand(tag);
	  				} else {                //Else add command.
	  					commands.push( parseTagCommand(tag) );
	  				}
					}
  			}
  			if(tag.tagName=="commandGroup") {
  				commandIndex = mozileMES.findByName(tag.getAttribute("name"), commands);
  				if(commandIndex >= 0) {
  					commands[commandIndex] = parseTagCommandGroup(tag);
  				} else {
  					commands.push( parseTagCommandGroup(tag) );
  				}
  			}
  		}
  		else {
  			if(tag.tagName=="command") {
  				commands.push( parseTagCommand(tag) );
  			}
  			if(tag.tagName=="commandGroup") {
  				commands.push( parseTagCommandGroup(tag) );
  			}
  		}
		}
	}

	tarr[3] = commands;
	
		// parse all the key tags into the keys array
	for(var k=0; k<keyTags.length; k++) {
		tag = keyTags[k];
		if ( tag.nodeType == Node.ELEMENT_NODE ) {
  		if(reuseKeys) {
  			if(tag.tagName=="key") {
  				//alert("Reuse KeyTag "+tag.getAttribute("name"));
  				keyIndex = findByName(tag.getAttribute("name"), keys);
  				if(keyIndex >= 0) {
  					//alert("Replace key at "+keyIndex);
  					keys[keyIndex] = parseTagKey(tag);
  				} else {
  					keys.push( parseTagKey(tag) );
  				}
  			}
  		} else {
  			if (tag.tagName=="key") {
  				//alert("New KeyTag "+tag.getAttribute("name"));
  				keys.push( parseTagKey(tag) );
  			}
  		}
		}
	}
	
	tarr[4] = keys;
	return tarr;
}

function parseTagCommand(tag) {
	var carr = new Array();
	carr[0] = tag.getAttribute("name");
	carr[1] = tag.getAttribute("accesskey");
	carr[2] = tag.getAttribute("button");
	if ( carr[0] == "separator" ) {
		carr[2] = "true";
	}
	return carr;
}

function parseTagCommandGroup(tag) {
	var garr = new Array();
	garr[0] = tag.getAttribute("name");
	garr[1] = tag.getAttribute("title");
	garr[2] = tag.getAttribute("tooltip");
	garr[3] = tag.getAttribute("icon");
	garr[4] = tag.getAttribute("accesskey");
	garr[5] = tag.getAttribute("button");

	var commands = new Array();
	var children = tag.childNodes;
	var i;
	for(i=0; i<children.length; i++) {
		if( (children[i].nodeType == Node.ELEMENT_NODE) && (children[i].tagName == "command") ) {
			commands.push( parseTagCommand(children[i]) );
		}
	}

	garr[6] = commands;
	
	return garr;
}

function parseTagKey(tag) {
	var karr = new Array();
	karr[0] = tag.getAttribute("name");
	karr[1] = tag.getAttribute("command");
	return karr;
}

	return tagArray;
} //end private functions


/**
Find the named command in an array, return the index.
*/
function mozileFindByName(name, array) {
	for(var i=0; i<array.length; i++) {
		if(array[i][0]==name) {
			return i;
		}
	}
		// nothing found
	return -1;
}

